# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .productinventorytype import ProductInventoryType


class BaseProductType(pydantic.BaseModel):
    display_name: str = pydantic.Field(
        default=...,
        title="Display name",
        alias='displayName',
        description=(
            "The human-readable name for this product type."
        )
    )

    inventory_type: ProductInventoryType = pydantic.Field(
        default=...,
        title="Inventory type",
        alias='inventoryType',
        description="Specifies the inventory type of the product."
    )

    is_final: bool = pydantic.Field(
        default=False,
        title="Is final?",
        alias='isFinal',
        description="Indicates if the product type is final and can not be subtyped."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import TypeVar

from .stringtype import StringType


T = TypeVar('T', bound='EuropeanVatNumber')


# https://gist.github.com/marcinlerka/630cc05d11bb10c5f1904506ff92abcd
class EuropeanVatNumber(StringType):
    patterns: re.Pattern[str] = re.compile(
        r'^('
        'ATU[0-9]{8}|'                              # Austria
        'BE0[0-9]{9}|'                              # Belgium
        'BG[0-9]{9,10}|'                            # Bulgaria
        'HR\\d{11}$|'                               # Croatia 
        'CY[0-5|9]\\d{7}[A-Z]|'                     # Cyprus
        'CZ[0-9]{8,10}|'                            # Czech Republic
        'DE[0-9]{9}|'                               # Germany
        'DK[0-9]{8}|'                               # Denmark
        'EE[0-9]{9}|'                               # Estonia
        '(EL|GR)[0-9]{9}|'                          # Greece
        'ES[0-9A-Z][0-9]{7}[0-9A-Z]|'               # Spain
        'EU\\d{9}|'                                 # EU-type 
        'FI[0-9]{8}|'                               # Finland
        'FR((?![IO])[A-Z0-9]){2}[0-9]{9}|'        # France
        'GB([0-9]{9}([0-9]{3})?|[A-Z]{2}[0-9]{3})|' # United Kingdom
        'HU[0-9]{8}|'                               # Hungary
        'IE[0-9A-Z\\*\\+]{7}[A-Z]{1,2}|'            # Ireland
        'IT[0-9]{11}|'                              # Italy
        'LT([0-9]{9}|[0-9]{12})|'                   # Lithuania
        'LU[0-9]{8}|'                               # Luxembourg
        'LV[0-9]{11}|'                              # Latvia
        'MT[0-9]{8}|'                               # Malta
        'NL[0-9]{9}B[0-9]{2}|'                      # Netherlands
        'PL[0-9]{10}|'                              # Poland
        'PT[0-9]{9}|'                               # Portugal
        'RO[0-9]{2,10}|'                            # Romania
        'SE[0-9]{12}|'                              # Sweden
        'SI[0-9]{8}|'                               # Slovenia
        'SK[0-9]{10}|'                              # Slovakia
        'NO\\d{9}|'                                 # Norway (not EU)
        'RU\\d{10}$|\\d{12}|'                       # Russia (not EU)
        'RS\\d{9}'                                  # Serbia (not EU)
        ')$'
    )

    @staticmethod
    def clean(v: str) -> str:
        return re.sub('[^0-9A-Z]', '', str.upper(v))

    @classmethod
    def is_valid(cls, vat: str) -> bool:
        return cls.patterns.match(vat) is not None
    
    @classmethod
    def validate_pattern(cls, v: str) -> str:
        return super().validate_pattern(cls.clean(v))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic

from canonical import DomainName
from canonical import EmailAddress
from canonical import ResourceName
from canonical.fields import ResourceVersionField
from .base import CompanyBasicInfoBase
from .companyaddress import CompanyAddress
from .companychaininfo import CompanyChainInfo
from .companyemailaddress import CompanyEmailAddress
from .companyphonenumber import CompanyPhonenumber


class CompanyBasicInfoV1Spec(pydantic.BaseModel):
    addresses: list[CompanyAddress] = pydantic.Field(
        default=[],
        title="Addresses",
        description="The list of known addresses."
    )

    company_name: str = pydantic.Field(
        default=...,
        title="Name",
        alias='companyName',
        description="The common name of the company."
    )

    chain: CompanyChainInfo = pydantic.Field(
        default_factory=CompanyChainInfo,
        titlte="Chain/franchise information",
        description=(
            "Describes the chain/franchise a business if part of."
        )
    )

    email_addresses: list[CompanyEmailAddress] = pydantic.Field(
        default=[],
        title="Email addresses",
        description=(
            "The list of known email addresses at which this company may "
            "be contacted."
        )
    )

    phonenumbers: list[CompanyPhonenumber] = pydantic.Field(
        default=[],
        title="Phonenumbers",
        description=(
            "The list of known phone numbers at which this company may "
            "be contacted."
        )
    )

    website: str | None = pydantic.Field(
        default=None,
        title="Website",
        description="The website of the business, if it has one."
    )

    class Config:
        allow_population_by_field_name: bool = True


class CompanyBasicInfoV1(CompanyBasicInfoBase):
    api_version: Literal['v1'] = ResourceVersionField('v1')
    spec: CompanyBasicInfoV1Spec

    def add_email(
        self,
        source: DomainName | ResourceName,
        email: EmailAddress,
        obtained: datetime.datetime | None = None
    ) -> None:
        self.spec.email_addresses.append(
            CompanyEmailAddress.parse_obj({
                'source': source,
                'value': email,
                'obtained': obtained or datetime.datetime.now(datetime.timezone.utc)
            })
        )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from ..resourcename import ResourceName
from ..resourcemetadata import ResourceMetadata
from .dto import DTO


class VersionedDTO(pydantic.BaseModel):
    __root__: DTO

    @property
    def metadata(self) -> ResourceMetadata:
        return self.__root__.metadata

    @property
    def name(self) -> ResourceName:
        return self.__root__.metadata.name

    @property
    def resource_id(self) -> str:
        return self.metadata.name.id

    @classmethod
    def parse(
        cls,
        *,
        name: ResourceName = ResourceName.null(),
        spec: dict[str, Any] | None = None,
        version: str | None = None,
        annotations: dict[str, str] | None = None,
        labels: dict[str, str] | None = None,
        tags: set[str] | None = None
    ):
        spec = spec or {}
        params: dict[str, Any] = {
            'kind': cls.__name__,
            'metadata': {
                'name': name,
                'annotations': annotations or {},
                'labels': labels or {},
                'tags': tags or set()
            },
            'spec': spec
        }
        if version is not None:
            params['apiVersion'] = version
        return cls.parse_obj(params)

    def annotate(self, name: str, value: Any, namespace: str | None = None) -> None:
        self.__root__.metadata.annotate(name, value, namespace)

    def assign(self, name: ResourceName) -> None:
        """Assign an identity to the DTO (by settings the :attr:`resource_id`)
        attribute.
        """
        self.metadata.name = name

    def get_label(self, name: str, namespace: str | None = None) -> Any:
        """Return the label in the optional namespace."""
        return self.metadata.get_label(name, namespace=namespace)

    def has_identity(self) -> bool:
        """Return a boolean indicating if the underlying resource has an
        identity.
        """
        return self.metadata.name != ResourceName.null()

    def is_annotated(self, name: str, namespace: str | None = None) -> bool:
        """Return a boolean indicating if the resource has the given
        annotation.
        """
        return self.__root__.metadata.is_annotated(name, namespace=namespace)

    def is_tagged(self, value: str | set[str]) -> bool:
        """Return a boolean indicating if the resource has the given
        tag.
        """
        return self.__root__.metadata.is_tagged(value)

    def label(self, name: str, value: Any, namespace: str | None = None) -> None:
        self.__root__.metadata.label(name, value, namespace)

    def tag(self, value: str) -> None:
        self.__root__.metadata.tag(value)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ProductInventoryType(str, enum.Enum):
    FINISHED_GOOD   = 'FINISHED_GOOD'
    MERCHANDIZE     = 'MERCHANDIZE'
    MRO             = 'MRO'
    NOOP            = 'NOT_APPLICABLE'
    RAW_MATERIAL    = 'RAW_MATERIAL'
    SUBASSEMBLY     = 'SUBASSEMBLY'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied
from typing import Any
from typing import Literal

import pydantic
import pydantic.main

from ..fields import ResourceVersionField
from ..resourcemetadata import ResourceMetadata
from .dtometaclass import DTOMetaclass


class DTO(pydantic.BaseModel, version='v1', metaclass=DTOMetaclass):
    """A Data Transfer Object (DTO) is an object that carries data between
    processes.
    """
    api_version: Literal['v1'] = ResourceVersionField('v1')
    kind: str
    metadata: ResourceMetadata = pydantic.Field(
        default=...,
        title="Metadata",
        description="Standard object's metadata."
    )
    spec: Any


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .productidentifier import ProductIdentifierType
from .productinventorytype import ProductInventoryType


class BaseProduct(pydantic.BaseModel):
    product_name: str = pydantic.Field(
        default=...,
        title="Name",
        description=(
            "The internal name for this Product. If there is no distinction "
            "between marketed names and internal names, this may also be "
            "the marketed name."
        )
    )

    identifiers: list[ProductIdentifierType] = pydantic.Field(
        default=[],
        title="Identifiers",
        description=(
            "The known identifiers for this product, such as stock keeping "
            "unit (SKU), ISBN, EAN, model number, etc."
        )
    )

    inventory_type: ProductInventoryType = pydantic.Field(
        default=...,
        title="Inventory type",
        alias='inventoryType',
        description=(
            "Specifies the inventory type of the product. Must be "
            "`NOT_APPLICABLE` if the product is a `SERVICE`."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic
import pydantic.main

from ..fields import ResourceKindField


class DTOMetaclass(pydantic.main.ModelMetaclass):

    def __new__(
        cls,
        name: str,
        bases: tuple[type[Any]],
        namespace: dict[str, Any],
        version: str,
        **params: Any
    ) -> 'DTOMetaclass':
        annotations: dict[str, type] = namespace.setdefault('__annotations__', {})
        if not namespace.pop('__abstract__', False):
            params.setdefault('allow_population_by_field_name', True)
            params.setdefault('title', str.title(version))
            if not namespace.get('Config'):
                namespace['Config'] = type('Config', (object,), params)
            annotations['kind'] = Literal[name] # type: ignore
            namespace['kind'] = ResourceKindField()

        new_cls = super().__new__(cls, name, bases, namespace) # type: ignore
        return new_cls # type: ignore


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .canonicalexception import CanonicalException
from .missingshippingdestination import MissingShippingDestination
from .ratelimited import RateLimited
from .staleresource import StaleResource
from .upstreamservicefailure import UpstreamServiceFailure


__all__: list[str] = [
    'CanonicalException',
    'MissingShippingDestination',
    'RateLimited',
    'StaleResource',
    'UpstreamServiceFailure',
]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ProductFeatureConstraint(str, enum.Enum):
    """Specifies how a feature (or a set of features) should be
    validated across all variants of a product.
    """
    #: No constraints are applied to the :term:`ProductFeature`.
    NONE                = 'NONE'

    #: Must be unique for its resource.
    UNIQUE              = 'UNIQUE'

    #: Must be unique a unique set of features.
    UNIQUE_COMBINATION  = 'UNIQUE_COMBINATION'

    #: Must be consistent.
    CONSISTENT          = 'CONSISTENT'



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied
import datetime
import urllib.parse
from typing import Any
from typing import Callable
from typing import TypeVar

import pydantic

from canonical import DomainName
from canonical import EmailAddress
from canonical import Phonenumber
from canonical import ResourceName
from .v1 import CompanyBasicInfoV1


T = TypeVar('T')


class CompanyBasicInfo(pydantic.BaseModel):
    """A Data Transfer Object (DTO) model to exchange basic information
    about companies.
    """
    __root__: CompanyBasicInfoV1

    @classmethod
    def parse(
        cls,
        *,
        name: ResourceName,
        spec: dict[str, Any],
        version: str | None = None,
        annotations: dict[str, str] | None = None,
        labels: dict[str, str] | None = None,
        tags: set[str] | None = None
    ):
        params: dict[str, Any] = {
            'kind': cls.__name__,
            'metadata': {
                'name': name,
                'annotations': annotations or {},
                'labels': labels or {},
                'tags': tags or set()
            },
            'spec': spec
        }
        if version is not None:
            params['apiVersion'] = version
        return cls.parse_obj(params)
    
    @property
    def domain(self) -> DomainName | None:
        """Return the domain name of the website, or ``None``."""
        if self.spec.website is not None:
            p = urllib.parse.urlparse(self.spec.website)
            return DomainName(p.netloc)

    @property
    def name(self) -> ResourceName:
        return self.__root__.metadata.name

    @property
    def phone(self) -> Phonenumber:
        phonenumbers = self.__root__.spec.phonenumbers
        if not phonenumbers:
            raise AttributeError(
            f"Company does not have a phonenumber (name: {self.name})"
            )
        if len(phonenumbers) > 1:
            raise AttributeError(
            f"Company has multiple phonenumbers (name: {self.name})"
            )
        return phonenumbers[0].value

    @property
    def spec(self):
        return self.__root__.spec

    def annotate(self, name: str, value: Any, namespace: str | None = None) -> None:
        qualname = name
        if namespace is not None:
            qualname = f'{namespace}/{name}'
        self.__root__.metadata.annotations[qualname] = value

    def get_annotation(
        self,
        name: str,
        parser: Callable[[str], T] = str,
        namespace: str | None = None
    ) -> None | T:
        """Get and parse the given annotation."""
        qualname = name
        if namespace is not None:
            qualname = f'{namespace}/{name}'
        value = self.__root__.metadata.annotations.get(qualname)
        if value is not None:
            value = parser(value)
        return value

    def add_email(
        self,
        source: DomainName | ResourceName,
        email: EmailAddress,
        obtained: datetime.datetime | None = None
    ) -> None:
        """Add an email address to this company."""
        if self.has_email(email):
            raise ValueError(f'Email {email} is already known for {self.name}')
        self.__root__.add_email(source, email, obtained)

    def get_tags(self) -> set[str]:
        """Return the set of tags that this resource was tagged with."""
        return self.__root__.metadata.tags

    def has_annotation(self, name: str, namespace: str | None = None) -> bool:
        """Return a boolean indicating if the resource has the given
        annotation.
        """
        qualname = name
        if namespace is not None:
            qualname = f'{namespace}/{name}'
        return qualname in self.__root__.metadata.annotations

    def has_email(self, email: EmailAddress | None = None) -> bool:
        """Return a boolean indicating if at least one email address
        is known for the company, if `email` is ``None``, otherwise
        return a boolean indicating if `email` is known.
        """
        if email is not None:
            return any([x.value == email for x in self.spec.email_addresses])
        return len(self.spec.email_addresses) > 0

    def has_phonenumber(self) -> bool:
        """Return a boolean indicating if at least one phone number
        is known for the company.
        """
        return len(self.spec.phonenumbers) > 0

    def has_website(self) -> bool:
        """Return a boolean indicating if the website is known for the company."""
        return bool(self.spec.website)

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .fields import ResourceAnnotationsField
from .resourcename import ResourceName


class ResourceMetadata(pydantic.BaseModel):
    name: ResourceName
    annotations: dict[str , str] = ResourceAnnotationsField()
    labels: dict[str, str] = {}
    tags: set[str] = set()

    @classmethod
    def _qualname(cls, namespace: str | None, name: str) -> str:
        qualname = name
        if namespace is not None:
            qualname = f'{namespace}/{name}'
        return qualname

    def annotate(self, name: str, value: Any, namespace: str | None = None) -> None:
        self.annotations[self._qualname(namespace, name)] = value
    
    def get_label(self, name: str, namespace: str | None = None) -> Any:
        """Return the label in the optional namespace."""
        return self.labels.get(self._qualname(namespace, name))

    def is_annotated(self, name: str, namespace: str | None = None) -> bool:
        """Return a boolean indicating if the resource has the given
        annotation.
        """
        return self._qualname(namespace, name) in self.annotations

    def is_tagged(self, value: str | set[str]) -> bool:
        """Return a boolean indicating if the resource has the given
        tag.
        """
        if isinstance(value, str):
            value = {value}
        return bool(value & self.tags)

    def label(self, name: str, value: Any, namespace: str | None = None) -> None:
        self.labels[self._qualname(namespace, name)] = value

    def tag(self, value: str) -> None:
        self.tags.add(value)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


class MissingShippingDestination(Exception):
    """Used in logistics applications to indicate that a shippable did not
    have a valid shipping destination attached to it.
    """
    __module__: str = 'canonical.exc'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import dto

from .v1 import Account as AccountV1


__all__: list[str] = [
    'Account',
    'AccountV1'
]


class Account(dto.VersionedDTO):
    __root__: AccountV1

    @property
    def spec(self):
        return self.__root__.spec


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic


__all__: list[str] = [
    'ResourceAnnotationsField',
    'ResourceKindField',
    'ResourceVersionField'
]


def ResourceVersionField(version: str) -> Any:
    return pydantic.Field(
        default=version,
        alias='apiVersion',
        title='API Version',
        description=(
            "The `apiVersion` property defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values. More info: https://git.k8s.io/community/"
            "contributors/devel/sig-architecture/api-conventions.md#resources"
        ),
        enum=[version]
    )


def ResourceKindField() -> Any:
    return pydantic.Field(
        default=...,
        title="Kind",
        description=(
            "Kind is a string value representing the conceptual model, entity or "
            "REST resource this object represents. Servers may infer this from "
            "the endpoint the client submits requests to. Cannot be updated. "
            "In `CamelCase`."
        )
    )


def ResourceAnnotationsField() -> Any:
    return pydantic.Field(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store and "
            "retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects. "
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical.fields import ResourceKindField
from canonical.resourcemetadata import ResourceMetadata


class CompanyBasicInfoBase(pydantic.BaseModel):
    kind: Literal['CompanyBasicInfo'] = ResourceKindField()
    metadata: ResourceMetadata = pydantic.Field(
        default=...,
        title="Metadata",
        description=(
            "Standard object's metadata. More info: https://git.k8s.io"
            "/community/contributors/devel/sig-architecture/api-conventions.md"
            "#metadata"
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic

from canonical import DomainName
from canonical import ResourceName
from canonical import Phonenumber
from canonical import URL


class CompanyPhonenumber(pydantic.BaseModel):
    kind: Literal['general'] | None = pydantic.Field(
        default=None,
        title="Kind",
        description=(
            "Specifies the kind of phonenumber, or `null` if no specific "
            "kind is known for this number."
        )
    )

    value: Phonenumber = pydantic.Field(
        default=...,
        title="Value",
        description=(
            "Contains the companies' phone number in international format. "
            "International format includes the country code, and is prefixed "
            "with the plus, +, sign. "
        )
    )

    source: URL | ResourceName | DomainName = pydantic.Field(
        default=DomainName('google.com'),
        title="Source",
        description=(
            "Indicates the source from which the phonenumber was obtained."
        )
    )

    obtained: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Obtained",
        description=(
            "Describes the date and time at which the phonenumber was "
            "obtained from the `source`."
        )
    )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical import GenericDeliveryPointSpecification
from canonical import GenericPostalAddress


class CompanyAddress(pydantic.BaseModel):
    kind: Literal['visiting', 'invoicing'] | None = pydantic.Field(
        default=None,
        title="Kind",
        description=(
            "Specifies the kind of address, or `null` if no specific "
            "kind is known for this address."
        )
    )

    spec: GenericPostalAddress | GenericDeliveryPointSpecification = pydantic.Field(
        default=...,
        title="Specification",
        description=(
            "Contains the companies' address as a `GenericPostalAddress` or "
            "`GenericDeliveryPointSpecification` object."
        )
    )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical import dto
from canonical import GenericDeliveryPointSpecification
from canonical import EuropeanVatNumber
from canonical.crm import OrganizationQualifiedContact


class AccountSpec(pydantic.BaseModel):
    organization_name: str = pydantic.Field(
        default=...,
        alias='organizationName',
        title="Organization name",
        description="The company or company division name.",
        max_length=63
    )

    organization_contact: OrganizationQualifiedContact = pydantic.Field(
        default=...,
        alias='organizationContact',
        title="Contact",
        description=(
            "Contact person within the organization that is responsible for this "
            "account."
        )
    )

    business_address: GenericDeliveryPointSpecification = pydantic.Field(
        default=...,
        alias='businessAddress',
        title="Business address",
        description="The account's business address information."
    )

    vat_number: EuropeanVatNumber | None = pydantic.Field(
        default=None,
        alias='vatNumber',
        title="VAT number",
        description=(
            "The VAT number as used in the European Union (EU) or "
            "the United Kingdom (UK)."
        )
    )

    class Config:
        allow_population_by_field_name: bool = True
        title = 'AccountSpecV1'


class Account(dto.DTO, version='v1'):
    spec: AccountSpec


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic

from canonical import DomainName
from canonical import ResourceName
from canonical import EmailAddress
from canonical import URL


class CompanyEmailAddress(pydantic.BaseModel):
    kind: Literal['general'] | None = pydantic.Field(
        default=None,
        title="Kind",
        description=(
            "Specifies the kind of email address, or `null` if no specific "
            "kind is known for this mailbox."
        )
    )

    value: EmailAddress = pydantic.Field(
        default=...,
        title="Value",
        description="Contains the companies' email address."
    )

    source: URL | ResourceName | DomainName = pydantic.Field(
        default=...,
        title="Source",
        description=(
            "Indicates the source from which the email address was obtained."
        )
    )

    obtained: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Obtained",
        description=(
            "Describes the date and time at which the email address was "
            "obtained from the `source`."
        )
    )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class FulfillmentMode(str, enum.Enum):
    #: Assemble to Order
    ATO = 'ATO'

    #: Configure to Order
    CTO = 'CTO'

    #: Engineer to Order
    ETO = 'ETO'

    #: Make to Order
    MTO = 'MTO'

    #: Make to Stock
    MTS = 'MTS'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ProductFeatureType(str, enum.Enum):
    #: Standard production feature; consistent accross all variants.
    STANDARD = 'STANDARD'

    #: Mandatory feature as part of the product definition. A
    #: required feature is consistent across all variants.
    REQUIRED = 'REQUIRED'

    #: The feature needs to be selected from a controlled
    #: vocabulary.
    SELECTABLE = 'SELECTABLE'

    #: The feature is optional.
    OPTIONAL = 'OPTIONAL'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import cast
from typing import TypeVar
import urllib.parse

from .domainname import DomainName
from .stringtype import StringType


T = TypeVar('T', bound='URL')


class URL(StringType):
    _parse_result: urllib.parse.ParseResult
    patterns: list[re.Pattern[str]] = [
        re.compile(r'(http|https)://.*')
    ]

    def __new__(
        cls: type[T],
        object: str,
        result: urllib.parse.ParseResult | None = None
    ) -> T:
        self = super().__new__(cls, object)
        if result is None:
            # Was not created by pydantic
            for validator in cls.__get_validators__():
                self = validator(self)
        else:
            assert result is not None
            self._parse_result = result
        return cast(T, self)
    
    @classmethod
    def validate(cls, v: str) -> str:
        p = urllib.parse.urlparse(v)
        if p.scheme not in {'http', 'https'}:
            raise ValueError(f"invalid protocol for URL: {p.scheme}.")
        for validator in DomainName.__get_validators__():
            validator(p.netloc)
        return cls(v, result=p)
    
    def __repr__(self) -> str:
        return f'URL({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical import ResourceName


class CompanyChainInfo(pydantic.BaseModel):
    is_branch: bool = pydantic.Field(
        default=False,
        alias='isBranch',
        title="Is branch?",
        description=(
            "Indicates if the company is part of a chain or franchise, "
            "and is a branch."
        )
    )

    parent: ResourceName | None = pydantic.Field(
        default=None,
        title="Parent",
        description=(
            "Identifies the chain or franchise headquarters."
        )
    )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical import EmailAddress
from canonical import Phonenumber


class OrganizationQualifiedContact(pydantic.BaseModel):
    """A reference to a contact inside an organization. At least the email
    address and the last name must be known.
    """
    first_name: str = pydantic.Field(
        default=...,
        alias='firstName',
        title="First name"
    )

    last_name: str = pydantic.Field(
        default=...,
        alias='lastName',
        title="Last name",
    )

    email: EmailAddress = pydantic.Field(
        default=...,
        title="Email address"
    )

    phonenumber: Phonenumber  = pydantic.Field(
        default=...,
        title="Phonenumber"
    )